/*
 * AVR based RGB Moodlight
 *
 * main.c
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 147 2012-01-11 22:05:20Z schwenke $
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <string.h>

#include "irmp.h"
#include "moodlight.h"


/* hardware definition  */

#define LEDPORT PORTC
#define LED_R	3	//PC3
#define LED_G	2	//PC2
#define LED_B	1	//PC1
#define LEDMASK (_BV(LED_R) | _BV(LED_G) | _BV(LED_B))

/* pwm data tables */

uint8_t pwm_next[4];
uint8_t pwm_data[4];
uint16_t pwm_time[4];

/* generated pwm table; factor=1.020000, mindelta=61 */

const uint16_t pwmtable[MAX_BRIGHT+1] PROGMEM = 
{
    0, 116, 177, 238, 299, 360, 421, 482, 543, 604, 665, 726, 787, 848,
    909, 970, 1031, 1092, 1153, 1214, 1275, 1336, 1397, 1458, 1519, 1580,
    1641, 1702, 1763, 1824, 1885, 1946, 2007, 2068, 2129, 2190, 2251, 2312,
    2373, 2434, 2495, 2556, 2617, 2678, 2739, 2800, 2861, 2922, 2983, 3044,
    3105, 3167, 3230, 3295, 3361, 3428, 3497, 3567, 3638, 3711, 3785, 3861,
    3938, 4017, 4097, 4179, 4262, 4348, 4435, 4523, 4614, 4706, 4800, 4896,
    4994, 5094, 5196, 5300, 5406, 5514, 5624, 5737, 5851, 5968, 6088, 6210,
    6334, 6460, 6590, 6721, 6856, 6993, 7133, 7275, 7421, 7569, 7721, 7875,
    8033, 8193, 8357, 8524, 8695, 8869, 9046, 9227, 9412, 9600, 9792, 9988,
    10187, 10391, 10599, 10811, 11027, 11248, 11473, 11702, 11936, 12175,
    12418, 12667, 12920, 13179, 13442, 13711, 13985, 14265, 14550, 14841,
    15138, 15441, 15750, 16065, 16386, 16714, 17048, 17389, 17737, 18091,
    18453, 18822, 19199, 19583, 19974, 20374, 20781, 21197, 21621, 22053,
    22494, 22944, 23403, 23871, 24348, 24835, 25332, 25839, 26356, 26883,
    27420, 27969, 28528, 29099, 29681, 30274, 30880, 31497, 32127, 32770,
    33425, 34094, 34776, 35471, 36181, 36904, 37642, 38395, 39163, 39946,
    40745, 41560, 42391, 43239, 44104, 44986, 45886, 46803, 47739, 48694,
    49668, 50661, 51675, 52708, 53762, 54838, 55934, 57053, 58194, 59358,
    60545, 61756, 62991, 64251, 0
};



/* initialize the hardware */
void hw_init(void)
{
    /* anything is input except LED lines */
    /* pullups enabled on all inputs */
    PORTB = 0xff;
    PORTC = ~LEDMASK;
    PORTD = 0xff;
    DDRB  = 0;
    DDRC  = LEDMASK;
    DDRD  = 0;

    /* disable Analog Comparator */
    ACSR = (_BV(ACD));

    /* timer1 runs @ sytem clk -> pwm=122Hz @ 8MHz */
    TCCR1B = _BV(CS10);

    /* timer2 counts in CTC mode 0..99 at clk/8
     * this gives 10.000 interrupts per second */
    TCCR2 = _BV(WGM21) | _BV(CS21);
    OCR2  = 99;

    /* timer1+2 compare match interrupt enabled */
    TIMSK = _BV(OCIE1A) | _BV(OCIE2);

    /* sleep enters IDLE mode */
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
}


/* the pwm ISR is time critical;
 * small steps require a fast ISR
 * that's why we bind the pwm phase
 * (the index in pwm tables) to r2
 */
register uint8_t pwm_phase asm("r2");

ISR(TIMER1_COMPA_vect)
{
    LEDPORT =   pwm_data[pwm_phase];
    pwm_phase = pwm_next[pwm_phase];
    if (pwm_phase == 0) --tick;
    OCR1A =     pwm_time[pwm_phase];
}


/* IR receiver sampling interrupt */
ISR(TIMER2_COMP_vect)
{
    sei();      /* allow the pwm ISR to interrupt us */
    irmp_ISR();
}


/* This is the main workhorse of the PWM
 * it translates the brighness of the 3 channels
 * to a table with LED bitmasks and timestamps
 * there are at most 4 compare matches in one timer cycle
 *
 * TCNT1 == 0  ... all LED w/ brightness > 0 turn on
 * TCNT1 == n1 ... the darkest LED turns off
 * TCNT1 == n2 ... the middle LED turns off
 * TCNT1 == n3 ... the brightest LED turns off
 *
 * LED with the same brightness will be combined to
 * a single entry in the table
 *
 * LED with brightness == 0 will never turn on
 * LED with brightness == MAX_BRIGHT will never turn off
 *
 * so by design we have n1 < n2 < n3
 */

void setup_pwm(void)
{
    /* initialize the data structure */
    memset(pwm_next, 0, 4*sizeof(uint8_t));
    memset(pwm_data, 0, 4*sizeof(uint8_t));
    memset(pwm_time, 0, 4*sizeof(uint16_t));

    /* write RGB brightness to PWM slots
     * no slot is used if LED is at 0% or 100%
     * slots are linked by the pwm_next[] array */
    {
        register uint8_t slot = 0;

        if (level_r != 0) 
		{
            pwm_data[0] |= _BV(LED_R);
            if (level_r < MAX_BRIGHT) 
			{
                /* link current slow to next one */
                pwm_next[slot] = slot+1;
                /* goto next slot */
                slot++;
                /* write LED + timestamp into next slot */
                pwm_data[slot] = _BV(LED_R);
                pwm_time[slot] = pgm_read_word(pwmtable + level_r);
            }
        }
        if (level_g != 0) 
		{
            pwm_data[0] |= _BV(LED_G);
            if (level_g < MAX_BRIGHT) 
			{
                pwm_next[slot] = slot+1;
                slot++;
                pwm_data[slot] = _BV(LED_G);
                pwm_time[slot] = pgm_read_word(pwmtable + level_g);
            }
        }
        if (level_b != 0) 
		{
            pwm_data[0] |= _BV(LED_B);
            if (level_b < MAX_BRIGHT) 
			{
                pwm_next[slot] = slot+1;
                slot++;
                pwm_data[slot] = _BV(LED_B);
                pwm_time[slot] = pgm_read_word(pwmtable + level_b);
            }
        }
        /* close loop: last slot points back to slot 0 */
        pwm_next[slot] = 0;
    }

    /* sort/merge the linked list of slots
     * this is bubblesort */
    {
        register uint8_t swapped;
        register uint8_t slot;
        register uint8_t next;
        register uint8_t secnext;

        do 
		{
            swapped = 0;
            slot = 0;
            while ((next = pwm_next[slot]) && (secnext = pwm_next[next])) 
			{
                if (pwm_time[next] == pwm_time[secnext]) 
				{
                    /* merge slot data */
                    pwm_data[next] |= pwm_data[secnext];
                    /* remove 'secnext' slot from list */
                    pwm_next[next] = pwm_next[secnext];
                    swapped = 1;
                }
                else if (pwm_time[next] > pwm_time[secnext]) 
				{
                    /* swap pointers */
                    pwm_next[next] = pwm_next[secnext];
                    pwm_next[slot] = secnext;
                    pwm_next[secnext] = next;
                    swapped = 1;
                }
                slot = pwm_next[slot];
            }
        } while (swapped);
    }

    /* finalize bitmasks; slot 0 turns LED(s) on
     * other slots turn respecive LED(s) off
     */
    {
        register uint8_t slot = 0;
        register uint8_t mask = pwm_data[0] | ~LEDMASK;
        pwm_data[0] = mask;
        while ((slot = pwm_next[slot])) 
		{
            mask &= ~pwm_data[slot];
            pwm_data[slot] = mask;
        }
    }
}


/* we have come a long way */

int __attribute__((OS_main)) main(void)
{

    hw_init();
    irmp_init();
    moodlight_init();
    setup_pwm();

    sei();

    while (1) 
	{
        while (tick) 
		{
            if (irmp_get_data(&irmp_data)) 
			{
                moodlight_remote();
            }
            sleep_cpu();
        }
        moodlight_step();
        setup_pwm();
    }

    return 0;
}
